library omnibus;

import 'dart:async';
import 'dart:collection';

/// Omnibus data
///
/// All data objects that is transferred trough the Omnibus have to use this
/// class as base class. You are have to extend this class and specify a [type].
/// All public data should be stored in the [data] field. This way other
/// languages than Dart will be able to connect to the Omnibus using a simple
/// mediator.
/// To distinguish between all the different data objects in the Omnibus, all
/// objects that use the same data structure and convey the same kind of data
/// have the same [type].
class OmnibusData
{
    /// Type identifier
    ///
    /// This is a unique identifier for the type of object that is contained in
    /// [data]. Type identifiers can follow any convention, as long as they are
    /// unique. It is recommended to define all type identifiers that are used
    /// in a specific package in one place to prevent collisions. Additionally,
    /// you can use the package name as prefix to prevent collisions with other
    /// packages. A recommended convention: `package.OriginEntity.originMethod`
    static final String type = 'default';

    String get instanceType => type;

    /// Data object
    ///
    /// Object containing the data that is being transferred. To connect the
    /// Omnibus to parts of the program that are written in another language,
    /// The [data] can be encoded into JSON.
    Object data;

    /// Create OmnibusData instance with the given [type] and [data].
    OmnibusData(this.data) { }
}

/// OmnibusEvent is an alias for OmnibusData and should be used as base class by
/// event classes.
class OmnibusEvent extends OmnibusData
{
    OmnibusEvent(Object data) : super(data);
}

/// OmnibusReply is an alias for OmnibusData and should be used as base class by
/// reply classes.
class OmnibusReply extends OmnibusData
{
    OmnibusReply(Object data) : super(data);
}

typedef bool                 OmnibusListenerProbe(OmnibusEvent event);
typedef Future<OmnibusReply> OmnibusListenerReceive(OmnibusEvent event);

/// Omnibus listener
///
/// Each instance of this class defines a separate listener. A listener is
/// connected to a channel in the Omnibus. You cannot connect the same listener
/// to multiple channels.
/// All incoming events are first probed using the [probe] function. If the
/// event is of any interest, [probe] returns true. The event is then passed on
/// to the [receive] function. The [receive] function can optionally send a
/// reply to the event sender by returning a non-null value.
abstract class OmnibusListener
{
    /// Tester to only receive a subset of all emitted events. This function
    /// should be very small.
    bool probe(OmnibusEvent event);

    /// Event handler which is called if [probe] returns true. This function
    /// should be programmed asynchronous.
    Future<OmnibusReply> receive(OmnibusEvent event);
}

/// Omnibus controller
///
/// The Omnibus controller contains a HashMap of listener sets. Each set
/// represents the listeners for a different [OmnibusData.type].
class Omnibus
{
    /// Key-value map for all data types.
    HashMap<String, Set<OminbusListener>> _listeners;

    /// Omnibus constructor
    Omnibus()
    {
        _listeners = new HashMap<String, Set<OminbusListener>>();
    }

    /// Add OminbusListener.
    void subscribe(String channel, OmnibusListener listener)
    {
        // Add new Set of listeners if absent for this OmnibusEvent.
        _listeners.putIfAbsent(channel, () => new Set<OminbusListener>());

        // Add this listener.
        _listeners[channel].add(listener);
    }

    /// Emit OmnibusEvent.
    Future<Set<OmnibusReply>> emit(
        OmnibusEvent event,
        [List<String> expect]) async
    {
        Completer completer = new Completer<Set<OmnibusReply>>();

        // Create Set for return value.
        Set<OmnibusReply> replies = new Set<OmnibusReply>();

        // Probe all listeners for this event.
        int listenerCount = _listeners[event.instanceType].length;
        _listeners[event.instanceType].forEach(
            (OmnibusListener listener)
            {
                // Proceed if the listener wants to receive this event.
                if (listener.probe(event))
                {
                    // Get reply from listener.
                    listener.receive(event).then(
                        (OmnibusReply reply)
                        {
                            replies.add(reply);
                            listenerCount--;
                            if (listenerCount == 0)
                            {
                                completer.complete(replies);
                            }
                        });
                }
            });

        // Return collected replies.
        return completer.future;
    }
}
